-----------------------------------------------------------------------------
--
-- data
--   \healingDone
--     \<healerName>
--       -count, amount, overhealAmount
--       \healDetails
--         \<healName>
--            -count, amount, overhealAmount
--			  \hitDetails
--			    \<healType> -- one of Heal, Critical Heal, Tick
--                -count, amount, overhealAmount
--   \healingTaken
--     \<healTargetName>
--       -count, amount, overhealAmount
--       \healDetails
--         \<healName>
--            -count, amount, overhealAmount
--			  \hitDetails
--			    \<healType> -- one of Heal, Critical Heal, Tick
--                -count, amount, overhealAmount
--
-----------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("AssessmentHealing")
L:RegisterTranslations("enUS", function() return {
	["Heal"] = true,
	["Overheal"] = true,
	["Healers List"] = true,
	["Healing Done Details"] = true,
	["Healing Done"] = true,
	["Fight Healing Done"] = true,
	["HPS"] = true,
	["Raw Healing Done"] = true,
	["Raw Healing Done Details"] = true,
	["Raw Healing Done Hit Details"] = true,
	["Fight Raw Healing Done"] = true,
	["Heal/Overheal"] = true,
	["Healing/Overhealing"] = true,
	["Healing Done Hit Details"] = true,

	["Healing"] = true,
	["Healing Options."] = true,
	["Colors"] = true,
	["Color options"] = true,
	["Healing"] = true,
	["Overhealing"] = true,

	["Tick"] = true,
	["Critical Heal"] = true,
	["Heal"] = true,

	["Heal Data"] = true,
	["Information about each heal used."] = true,
	["Hit Data"] = true,
	["Detailed information about each attack used."] = true,

	["FILTER_NONE"]	= "Gather no data.",
	["FILTER_SELF"] = "Gather self data only.",
	["FILTER_ALL"]	= "Gather all data.",
} end)

-----------------------------------------------------------------------------

local AssessmentHealing = Assessment:NewModule("Healing", "AceEvent-2.0", "Parser-3.0")
AssessmentHealing.db = Assessment:AcquireDBNamespace("Healing")
Assessment:RegisterDefaults("Healing", "profile", {
	enable					= true,
	gatherHealDataFilter	= "ALL",
	gatherHitDataFilter		= "ALL",
	healColor				= { 1, 1, 1, 0 },
	overhealColor			= { 0, 0, 0, 0.5 },
})

local profile
local gatherHealDataMethod
local gatherHitDataMethod

local emptyMethod = Assessment.EmptyMethod

-----------------------------------------------------------------------------

Assessment.TagData.healingdone =
{
	method = function(data)
				local fightData = data.fightData
				if not fightData then return 0 end
				
				local healingDone = fightData.healingDone
				if not healingDone then return 0 end
				
				if not healingDone[data.key] then return 0 end
				return healingDone[data.key].amount
			 end,
	data = "healing",
	guaranteeResult = true,
	numeric = true,
}

Assessment.TagData.overhealingdone =
{
	method = function(data)
				local fightData = data.fightData
				if not fightData then return 0 end
				
				local healingDone = fightData.healingDone
				if not healingDone then return 0 end
				
				if not healingDone[data.key] then return 0 end
				return healingDone[data.key].overhealAmount
			 end,
	data = "healing",
	guaranteeResult = true,
	numeric = true,
}

Assessment.TagData.healingtaken =
{
	method = function(data)
				local fightData = data.fightData
				if not fightData then return 0 end
				
				local healingTaken = fightData.healingTaken
				if not healingTaken then return 0 end
				
				if not healingTaken[data.key] then return 0 end
				return healingTaken[data.key].amount
			 end,
	data = "healing",
	guaranteeResult = true,
	numeric = true,
}

Assessment.TagData.overhealingtaken =
{
	method = function(data)
				local fightData = data.fightData
				if not fightData then return 0 end
				
				local healingTaken = fightData.healingTaken
				if not healingTaken then return 0 end
				
				if not healingTaken[data.key] then return 0 end
				return healingTaken[data.key].overhealAmount
			 end,
	data = "healing",
	guaranteeResult = true,
	numeric = true,
}

Assessment.TagData.overhealamount =
{
	tableData = "value.overhealAmount",
	data = "healing",
	numeric = true,
}

-----------------------------------------------------------------------------

Assessment.ListType.EnumerateTable.updateData.healingDone = "healing"
Assessment.ListType.EnumerateTable.updateData.healingTaken = "healing"

-----------------------------------------------------------------------------

local playerName = UnitName("player")
local gatherFilterMethods =
{
	NONE	= function() return false end,
	SELF	= function(name) return name == playerName end,
	ALL		= function() return true end,
}

-----------------------------------------------------------------------------

Assessment.ListType.Healers =
{
	method = function(fightData)
				if not fightData.healingDone or not fightData.healingDone then return emptyMethod end
				return pairs(fightData.healingDone)
			 end,
	data = "healing",
	description = L["Healers List"],
}

local function HealOverhealIterator(state, index)
	if not index then 
		return L["Heal"], state.amount-state.overhealAmount
	elseif index == L["Heal"] then
		return L["Overheal"], state.overhealAmount
	end
end

Assessment.ListType.HealOverheal =
{
	method = function(fightData, ...)
				local data = fightData.healingDone
				if not data then return emptyMethod end

				for i=1,select('#',...) do
					if not data then return emptyMethod end
					data = data[select(i, ...)]
				end
				if not data then return emptyMethod end
				
				return HealOverhealIterator, data, nil
			 end,

	data = "healing",
	description = L["Heal/Overheal"]
}

-----------------------------------------------------------------------------

local keyToColorNameMap = { [L["Heal"]] = "healColor", [L["Overheal"]] = "overhealColor" }
Assessment.ColorType.HealOverheal = 
{
	method = function(fightData, data)
				local colorName = keyToColorNameMap[data.key]
				if not colorName then return nil end
				return profile[colorName]
			 end,
	data = "healing",
	description = L["Heal/Overheal"]
}

-----------------------------------------------------------------------------

Assessment.DefaultDisplaySets.HealingDoneHitDetails =
{
	titleTag			= "[windowParameter(1)]: [windowParameter(2)]",
	listType			= "EnumerateTable", 
	listDataTag			= "healingDone,[windowParameter(1)],healDetails,[windowParameter(2)],hitDetails",
	barValueTag			= "amount - overhealAmount",
	filterTag			= "1",
	fightIndex			= -1,
	text				= { { tag = "[rank]. [key] ([count])" }, { tag = "[barValue:short]" } },
	description			= L["Healing Done Hit Details"],
	selectable			= false,
}

Assessment.DefaultDisplaySets.HealingDoneDetails =
{
	titleTag			= "[(fightIndex ~= 0):\"Fight \"]Healing Done: [windowParameter(1)]",
	listType			= "EnumerateTable", 
	listDataTag			= "healingDone,[windowParameter(1)],healDetails",
	barValueTag			= "amount - overhealAmount",
	filterTag			= "1",
	fightIndex			= -1,
	text				= { { tag = "[rank]. [key] ([count])" }, { tag = "[barValue:short]" } },
	description			= L["Healing Done Details"],
	selectable			= false,
	onClickType			= "SHOW_DISPLAY_SET",
	onClickDisplaySet	= L["Healing Done Hit Details"],
	onClickData			= "[windowParameter(1)],[key]",
}

Assessment.DefaultDisplaySets.HealingDone =
{
	titleTag			= "Healing Done: [totalBarValue:short]",
	listType			= "GroupMembers", 
	barValueTag			= "healingDone - overHealingDone",
	filterTag			= "barValue ~= 0",
	fightIndex			= 0,
	text				= { { }, { tag= "[barValue:short]" }, { tag = "[(barValue*100/totalBarValue):format(\"%.1f%%\")]", visible = false }, count = 3 },
	description			= L["Healing Done"],
	onClickType			= "SHOW_DISPLAY_SET",
	onClickDisplaySet	= L["Healing Done Details"],
	onClickData			= "[key]",
	colorTypeList		= { "Class", count = 1 },
}

Assessment.DefaultDisplaySets.FightHealingDone =
{
	titleTag			= "Fight Healing Done: [totalBarValue:short]",
	listType			= "GroupMembers", 
	barValueTag			= "healingDone - overhealingDone", 
	filterTag			= "barValue ~= 0",
	fightIndex			= 1,
	text				= { { }, { tag= "[barValue:short]" }, { tag = "[(barValue*100/totalBarValue):format(\"%.1f%%\")]", visible = false }, count = 3 },
	description			= L["Fight Healing Done"],
	onClickType			= "SHOW_DISPLAY_SET",
	onClickDisplaySet	= L["Healing Done Details"],
	onClickData			= "[key]",
	colorTypeList		= { "Class", count = 1 },
}

Assessment.DefaultDisplaySets.HPS =
{
	titleTag			= "HPS: [combatDuration:formatDuration]",
	listType			= "GroupMembers", 
	barValueTag			= "(healingDone - overhealingDone) / maximum(combatDuration, 1)", 
	filterTag			= "barValue ~= 0",
	fightIndex			= 1,
	text				= { { }, { tag= "[barValue:format(\"%.1f\")]" }, { tag = "[(barValue*100/totalBarValue):format(\"%.1f%%\")]", visible = false }, count = 3 },
	description			= L["HPS"],
	onClickType			= "SHOW_DISPLAY_SET",
	onClickDisplaySet	= L["Healing Done Details"],
	onClickData			= "[key]",
	favorite			= false,
	colorTypeList		= { "Class", count = 1 },
}

Assessment.DefaultDisplaySets.HealingOverhealing =
{
	titleTag			= "[windowParameter(1)]",
	listType			= "HealOverheal", 
	listDataTag			= "[windowParameterList]",
	barValueTag			= "value", 
	filterTag			= "1",
	fightIndex			= -1,
	text				= { { }, { tag= "[barValue:short]" } },
	description			= L["Healing/Overhealing"],
	colorTypeList		= { "HealOverheal", count = 1 },
	sortKeyTag			= "rank",
	descendingSort		= false,
	favorite			= false,
	selectable			= false,
}

Assessment.DefaultDisplaySets.RawHealingDoneHitDetails =
{
	titleTag			= "[windowParameter(1)]: [windowParameter(2)]",
	listType			= "EnumerateTable", 
	listDataTag			= "healingDone,[windowParameter(1)],healDetails,[windowParameter(2)],hitDetails",
	barValueTag			= "amount",
	filterTag			= "1",
	fightIndex			= -1,
	text				= { { }, { tag = "[barValue:short][isReport:\" Raw\"]" }, { tag = "[(amount-overhealAmount):short] Actual, [overhealAmount:short] Overheal ([(overhealAmount*100/amount):format(\"%.1f%%\")])", visible = false }, count = 3 },
	description			= L["Raw Healing Done Hit Details"],
	selectable			= false,
	segmentEnable		= true,
	segmentDisplaySet	= L["Healing/Overhealing"],
	segmentData			= "[windowParameter(1)],healDetails,[key]",
}

Assessment.DefaultDisplaySets.RawHealingDoneDetails =
{
	titleTag			= "[(fightIndex ~= 0):\"Fight \"]Raw Healing Done: [windowParameter(1)]",
	listType			= "EnumerateTable", 
	listDataTag			= "healingDone,[windowParameter(1)],healDetails",
	barValueTag			= "amount",
	filterTag			= "1",
	fightIndex			= -1,
	text				= { { }, { tag = "[barValue:short][isReport:\" Raw\"]" }, { tag = "[(amount-overhealAmount):short] Actual, [overhealAmount:short] Overheal ([(overhealAmount*100/amount):format(\"%.1f%%\")])", visible = false }, count = 3 },
	description			= L["Raw Healing Done Details"],
	selectable			= false,
	onClickType			= "SHOW_DISPLAY_SET",
	onClickDisplaySet	= L["Raw Healing Done Hit Details"],
	onClickData			= "[windowParameter(1)],[key]",
	segmentEnable		= true,
	segmentDisplaySet	= L["Healing/Overhealing"],
	segmentData			= "[windowParameter(1)],healDetails,[key]",
}

Assessment.DefaultDisplaySets.FightRawHealingDone =
{
	titleTag			= "Fight Raw Healing Done: [totalBarValue:short]",
	listType			= "GroupMembers", 
	barValueTag			= "healingDone", 
	filterTag			= "barValue ~= 0",
	fightIndex			= 1,
	text				= { { }, { tag = "[barValue:short][isReport:\" Raw\"]" }, { tag = "[(healingDone-overhealingDone):short] Actual, [overhealingDone:short] Overheal ([(overhealingDone*100/healingDone):format(\"%.1f%%\")])", visible = false }, count = 3 },
	description			= L["Fight Raw Healing Done"],
	favorite			= false,
	onClickType			= "SHOW_DISPLAY_SET",
	onClickDisplaySet	= L["Raw Healing Done Details"],
	onClickData			= "[key]",
	segmentEnable		= true,
	segmentDisplaySet	= L["Healing/Overhealing"],
	segmentData			= "[key]",
	colorTypeList		= { "Class", count = 1 },
}

Assessment.DefaultDisplaySets.RawHealingDone =
{
	titleTag			= "Raw Healing: [totalBarValue:short]",
	listType			= "GroupMembers", 
	barValueTag			= "healingDone", 
	filterTag			= "barValue ~= 0",
	fightIndex			= 0,
	text				= { { }, { tag = "[barValue:short][isReport:\" Raw\"]" }, { tag = "[(healingDone-overhealingDone):short] Actual, [overhealingDone:short] Overheal ([(overhealingDone*100/healingDone):format(\"%.1f%%\")])", visible = false }, count = 3 },
	description			= L["Raw Healing Done"],
	onClickType			= "SHOW_DISPLAY_SET",
	onClickDisplaySet	= L["Raw Healing Done Details"],
	onClickData			= "[key]",
	segmentEnable		= true,
	segmentDisplaySet	= L["Healing/Overhealing"],
	segmentData			= "[key]",
	colorTypeList		= { "Class", count = 1 },
}

-----------------------------------------------------------------------------

Assessment.Options.args.data.args.healing =
{
	type = "group",
	name = L["Healing"],
	desc = L["Healing Options."],
	args =
	{
		gatherHealData =
		{
			type = "text",
			name = L["Heal Data"],
			desc = L["Information about each heal used."],
			get = function() return profile.gatherHealDataFilter end,
			set = function(v) profile.gatherHealDataFilter = v AssessmentHealing:OnInitialize() end,
			validate = { NONE = L["FILTER_NONE"], SELF = L["FILTER_SELF"], ALL = L["FILTER_ALL"] }
		},
		gatherHitData =
		{
			type = "text",
			name = L["Hit Data"],
			desc = L["Detailed information about each attack used."],
			get = function() return profile.gatherHitDataFilter end,
			set = function(v) profile.gatherHitDataFilter = v AssessmentHealing:OnInitialize() end,
			validate = { NONE = L["FILTER_NONE"], SELF = L["FILTER_SELF"], ALL = L["FILTER_ALL"] }
		},
		color =
		{
			type = "group",
			name = L["Colors"],
			desc = L["Color options"],
			args =
			{
				healing =
				{
					type = "color",
					name = L["Healing"],
					desc = L["Healing"],
					get = function() return unpack(AssessmentHealing.db.profile.healColor) end,
					set = function(r, g, b, a) AssessmentHealing.db.profile.healColor = { r, g, b, a } Assessment:RedrawAll() end,
					hasAlpha = true
				},
				overHealing =
				{
					type = "color",
					name = L["Overhealing"],
					desc = L["Overhealing"],
					get = function() return unpack(AssessmentHealing.db.profile.overhealColor) end,
					set = function(r, g, b, a) AssessmentHealing.db.profile.overhealColor = { r, g, b, a } Assessment:RedrawAll() end,
					hasAlpha = true
				},
			},
		},
	}
}

-----------------------------------------------------------------------------

function AssessmentHealing:OnInitialize()
	profile = self.db.profile
	gatherHealDataMethod = gatherFilterMethods[profile.gatherHealDataFilter]
	gatherHitDataMethod = gatherFilterMethods[profile.gatherHitDataFilter]
end

function AssessmentHealing:OnEnable()
	self:RegisterParserEvent({ eventType = "Heal" }, "OnParserHealEvent")
end

function AssessmentHealing:OnDisable()
end

-----------------------------------------------------------------------------

local function AccumulateData(data, eventData)
	data.count = data.count + 1
	data.amount = data.amount + eventData.amount
	data.overhealAmount = data.overhealAmount + eventData.overhealAmount
end

-----------------------------------------------------------------------------

function AssessmentHealing:NewDataTable()
	local data = self:NewTable()
	data.count = 0
	data.amount = 0
	data.overhealAmount = 0
	return data
end

-----------------------------------------------------------------------------

function AssessmentHealing:UpdateHealingData(dataTable, member, eventData, healType)
	local data = dataTable[member]
	if not data then
		data = self:NewDataTable()
		data.healDetails = self:NewTable()
		dataTable[member] = data
	end
	AccumulateData(data, eventData)
	
	if gatherHealDataMethod(member) then
		local healDetailsParent = data.healDetails
		local healDetails = healDetailsParent[eventData.abilityName]
		if not healDetails then
			healDetails = self:NewDataTable()
			healDetails.hitDetails = self:NewTable()
			healDetailsParent[eventData.abilityName] = healDetails
		end
		AccumulateData(healDetails, eventData)

		if gatherHitDataMethod(member) then
			local healTypeDetailsParent = healDetails.hitDetails
			local healTypeDetails = healTypeDetailsParent[healType]
			if not healTypeDetails then
				healTypeDetails = self:NewDataTable()
				healTypeDetailsParent[healType] = healTypeDetails
			end
			AccumulateData(healTypeDetails, eventData)
		end
	end
end

-----------------------------------------------------------------------------

function AssessmentHealing:OnParserHealEvent(eventData)
	local healType
	if eventData.isDoT then
		healType = L["Tick"]
	elseif eventData.isCrit then
		healType = L["Critical Heal"]
	else
		healType = L["Heal"]
	end
	
	for data in self:IterateUpdateDataTables() do
		data.dataUpdateTimes.healing = GetTime()
		local healingDone = data.healingDone
		if not healingDone then
			healingDone = self:NewTable()
			data.healingDone = healingDone
		end

		local healingTaken = data.healingTaken
		if not healingTaken then
			healingTaken = self:NewTable()
			data.healingTaken = healingTaken
		end

		self:UpdateHealingData(healingDone, eventData.sourceName, eventData, healType)
		self:UpdateHealingData(healingTaken, eventData.recipientName, eventData, healType)
	end
end

-----------------------------------------------------------------------------
