-----------------------------------------------------------------------------
--
-- data
--  \threat
--    \<playerName>
--      - amount
--		\targetDetails
--        <targetHash> : amount
--
-----------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("AssessmentThreat")
L:RegisterTranslations("enUS", function() return {
	["Threat List"] = true,
	["Threat Deficit"] = true,
	["Aggro Gain"] = true,
	["Aggro Gain Color"] = true,
	["Threat"] = true,
	["Threat gap before aggro gain on target"] = true,

	["Single Target Mode"] = true,
	["Single target mode causes Assessment to clear threat on mobs of that name when they die."] = true,

	["Threat"] = true,
	["Threat Options."] = true,
	["Aggro Gain Color"] = true,
	["Aggro gain color"] = true,
	
	["Group Threat Versions"] = true,

	["Aggro Gain Text"] = true,
	["Text to show for the aggro gain bar."] = true,
	["Any Text"] = true,

	["Enable KLHTM Broadcast"] = true,
	["Broadcast threat data to KLHThreatMeter users."] = true,
} end)

local AssessmentThreat = Assessment:NewModule("Threat", "AceEvent-2.0", "Parser-3.0")
AssessmentThreat.db = Assessment:AcquireDBNamespace("Threat")
Assessment:RegisterDefaults("Threat", "profile", {
	enable				= true,
	singleTargetMode	= false,
	aggroGainColor		= { 1, 0, 0 },
	aggroGainText		= L["Aggro Gain"],
	klhtmBroadcast		= false,
})

local Threat = AceLibrary("Threat-1.0")
local THREAT_GLOBAL_TARGET = Threat.GlobalTarget

-----------------------------------------------------------------------------

local aggroGainText
local profile

local emptyMethod = Assessment.EmptyMethod

-----------------------------------------------------------------------------

Assessment.TagData.targetthreat =
{
	method = function(data)
				local targetName = UnitName("target")
				if not targetName then return 0 end
				if data.key == aggroGainText then
					return Assessment.TagData.aggrogaintargetthreat.method(data)
				else
					return Threat:GetThreat(data.key, targetName)
				end
			 end,
	data			= { "target", "threat" },
	guaranteeResult = true,
	numeric			= true
}

Assessment.TagData.maxtargetthreat =
{
	method = function(data)
				local targetName = UnitName("target")
				if not targetName then return 0 end
				return Threat:GetMaxThreatOnTarget(targetName)
			 end,
	data			= { "target", "threat" },
	guaranteeResult = true,
	numeric			= true
}

Assessment.TagData.aggrogaintargetthreat =
{
	method = function(data)
				local targetName = UnitName("target")
				if not targetName then return 0 end

				if UnitIsUnit("player", "targettarget") then
					return Threat:GetThreat(UnitName("player"), targetName)
				end
				
				local multiplier = Threat:UnitInMeleeRange("target") and 1.1 or 1.3
				local targetTargetName = UnitName("targettarget")
				if not targetTargetName then
					return Threat:GetMaxThreatOnTarget(targetName)
				end
				return Threat:GetThreat(targetTargetName, targetName) * multiplier
			 end,
	data			= { "target", "threat" },
	guaranteeResult = true,
	numeric			= true
}

Assessment.TagData.threat =
{
	method = function(data)
				if data.key == aggroGainText then return 0 end

				local fightData = data.fightData
				if not fightData then return 0 end
				
				local threat = fightData.threat
				if not threat then return 0 end
				
				if not threat[data.key] then return 0 end
				return threat[data.key].amount					
			 end,
	data			= { "threat" },
	guaranteeResult = true,
	numeric			= true
}

Assessment.TagData.threatversion =
{
	method = function(data, prior)
				if not prior then return end
				if data.key == UnitName("player") then
					local _, v = Threat:GetLibraryVersion()
					return v
				else
					return tonumber(Threat:GetPartyRevisions()[data.key])
				end
			 end,
	numeric = true,
}

Assessment.TagData.maxthreat =
{
	method = function(data)
				local fightData = data.fightData
				if not fightData then return 0 end
				
				local threat = fightData.threat
				if not threat then return 0 end

				local maxThreat = 0
				for k,v in pairs(threat) do
					if v.amount > maxThreat then
						maxThreat = v.amount
					end
				end
				return maxThreat
			 end,
	data			= { "threat" },
	guaranteeResult = true,
	numeric			= true
}

Assessment.TagData.threattargetname =
{
	method = function(data) 
		if UnitExists("target") and not UnitIsFriend("player", "target") then
			return UnitName("target")
		end
	end,
}

Assessment.TagData.isusingforeignthreatsource =
{
	method = function(data) return Threat:IsUsingForeignThreatSource(data.key) end,
}

-----------------------------------------------------------------------------

local function ThreatListIterator(state, index)
	if not index then
		return aggroGainText
	elseif index == aggroGainText then
		return next(state)
	else
		return next(state, index)
	end
end

Assessment.ListType.ThreatList =
{
    method = function(fightData)
                if not fightData.threat then return emptyMethod end
				if UnitExists("targettarget") and not UnitIsFriend("player", "target") and not UnitIsUnit("player", "targettarget") then
					return ThreatListIterator, fightData.threat, nil
				end
				return pairs(fightData.threat)
             end,
    data = "threat",
    description = L["Threat List"],
}

-----------------------------------------------------------------------------

Assessment.ColorType.AggroGain =
{
	method = function(fightData, data)
				if data.key ~= aggroGainText then return nil end
				return profile.aggroGainColor
			 end,
	description = L["Aggro Gain Color"]
}

-----------------------------------------------------------------------------

Assessment.DefaultDisplaySets.Threat =
{
	titleTag		= "[threatTargetName:yellow][~threatTargetName:\"Threat\"]",
	listType		= "ThreatList", 
	barValueTag		= "select(threatTargetName, targetthreat, threat)", 
	filterTag		= "barValue ~= 0",
	fightIndex		= 0,
	text			= { 
						{ tag = "[rank]. [key]", offsets = { right = -45 } }, 
						{ tag = "[isUsingForeignThreatSource:\"*\"][barValue:short]" },
						{ tag = "[(select(threatTargetName, targetthreat, threat)/maximum(currentCombatDuration, 1)):format(\"%.1f\")]", alignment = "RIGHT", offsets = { right = -25 } },
						{ tag = "[(select(threatTargetName, targetthreat/maxTargetThreat, threat/maxThreat)*100):percent]", alignment = "RIGHT", visible = false, offsets = { right = -50 }  },
						count = 4,
					  },
	description		= L["Threat"],
	colorTypeList	= { "AggroGain", "Player", "Class", count = 3 },
}

Assessment.DefaultDisplaySets.GroupThreatVersions =
{
	titleTag		= "Threat Versions [threatVersion:prefix(\"r\"):yellow]",
	listType		= "CurrentGroupMembers", 
	barValueTag		= "threatVersion",
	sortKeyTag		= "barValue",
	filterTag		= "1",
	fightIndex		= 0,
	text			= { { tag = "[key]" }, { tag= "[select(threatVersion, threatVersion:prefix(\"r\"):yellow, \"Unknown\":red)]",  } },
	description		= L["Group Threat Versions"],
	colorTypeList	= { "Class", count = 1 },
	favorite		= false,
}

-----------------------------------------------------------------------------

Assessment.DefaultTags.ThreatDeficit =
{
	tag			= "[(maxTargetThreat-targetThreat):short]",
	fightIndex	= 0,
	description	= L["Threat Deficit"]
}

Assessment.DefaultTags.TargetAggroThreatDeficit =
{
	tag			= "[(aggrogaintargetthreat-targetThreat):short]",
	fightIndex	= 0,
	description	= L["Threat gap before aggro gain on target"]
}

-----------------------------------------------------------------------------

Assessment.Options.args.data.args.threat =
{
	type = "group",
	name = L["Threat"],
	desc = L["Threat Options."],
	args =
	{
		aggroGainColor =
		{
			type = "color",
			name = L["Aggro Gain Color"],
			desc = L["Aggro gain color"],
			get = function() return unpack(profile.aggroGainColor) end,
			set = function(r, g, b) profile.aggroGainColor = { r, g, b, a } Assessment:RedrawAll() end,
		},
		singleTargetMode =
		{
			type = "toggle",
			name = L["Single Target Mode"],
			desc = L["Single target mode causes Assessment to clear threat on mobs of that name when they die."],
			get = function() return profile.singleTargetMode end,
			set = function(v) AssessmentThreat:SetSingleTargetMode(v) end,
		},
		aggroGainText =
		{
			type = "text",
			name = L["Aggro Gain Text"],
			desc = L["Text to show for the aggro gain bar."],
			get = function() return profile.aggroGainText end,
			set = function(v) profile.aggroGainText = v aggroGainText = v Assessment:RedrawAll() end,
			usage = L["Any Text"],
		},
		klhtmBroadcast =
		{
			type = "toggle",
			name = L["Enable KLHTM Broadcast"],
			desc = L["Broadcast threat data to KLHThreatMeter users."],
			get = function() return Threat:IsKLHTMBroadcastEnabled() end,
			set = function(v) profile.klhtmBroadcast = v Threat:EnableKLHTMBroadcast(v) end,
		},
	}
}

-----------------------------------------------------------------------------

function AssessmentThreat:OnInitialize()
	profile = self.db.profile
	aggroGainText = profile.aggroGainText
end

function AssessmentThreat:OnEnable()
	self:RegisterEvent("PLAYER_TARGET_CHANGED", "OnPlayerTargetChanged")
	self:RegisterEvent("Threat_ThreatUpdated", "OnThreatUpdated")
	self:RegisterEvent("Threat_ThreatCleared", "OnThreatCleared")
	
	if profile.singleTargetMode then
		self:RegisterParserEvent({ eventType = "Death" }, "OnParserDeathEvent")
	end

	Threat:EnableKLHTMBroadcast(profile.klhtmBroadcast)
end

function AssessmentThreat:OnDisable()
end

-----------------------------------------------------------------------------

function AssessmentThreat:SetSingleTargetMode(v)
	profile.singleTargetMode = v
	
	if v then
		self:RegisterParserEvent({ eventType = "Death" }, "OnParserDeathEvent")
	else
		self:UnregisterAllParserEvents()
	end
end

function AssessmentThreat:OnPlayerTargetChanged()
	for data in self:ForceIterateUpdateDataTables() do
		data.dataUpdateTimes.target = GetTime()
	end
end

function AssessmentThreat:OnThreatCleared()
	local data = Assessment:GetFightData(1)
	if data and data.threat then
		data.threatLocked = true
	end
	
	data = Assessment:GetFightData(0)
	if data.threat then
		data.dataUpdateTimes.threat = GetTime()
		data.threat = self:ReleaseTable(data.threat, true)
	end
end

function AssessmentThreat:OnThreatUpdated(playerName, playerUnit, targetName)
	local threatValue = Threat:GetThreat(playerName, targetName)
	for data in self:ForceIterateUpdateDataTables() do
		data.dataUpdateTimes.threat = GetTime()
		
		local threat = data.threat
		if not threat then
			threat = self:NewTable()
			data.threat = threat
		end
		
		local playerData = threat[playerName]
		if playerData then
			if targetName == THREAT_GLOBAL_TARGET and threatValue == 0 then
				if not data.threatLocked then
					threat[playerName] = self:ReleaseTable(playerData, true)
				end
			else
				if threatValue == 0 then
					playerData.targetDetails[targetName] = nil
				else
					playerData.targetDetails[targetName] = threatValue
				end
				
				if threatValue >= playerData.amount then
					playerData.amount = threatValue
				else
					local maximumThreatValue = threatValue
				
					for k,v in pairs(playerData.targetDetails) do
						if v > maximumThreatValue then
							maximumThreatValue = v
						end
					end
					
					if maximumThreatValue == 0 then
						threat[playerName] = self:ReleaseTable(playerData, true)
					else
						playerData.amount = maximumThreatValue
					end
				end
			end
		elseif threatValue ~= 0 then
			playerData = self:NewTable()
			playerData.amount = threatValue
			playerData.targetDetails = self:NewTable()
			playerData.targetDetails[targetName] = threatValue
			threat[playerName] = playerData
		end
	end
end

function AssessmentThreat:OnParserDeathEvent(eventData)
	-- Only clear threat on things killed by yourself, or someone in your group.
	if not eventData.sourceID or not UnitAffectingCombat("player") then return end

	Threat:GetPlayerModule():_setTargetThreat(0, eventData.recipientName)
	if UnitExists("pet") then
		Threat:GetPetModule():_setTargetThreat(0, eventData.recipientName)
	end
end

-----------------------------------------------------------------------------
